/*
 * Copyright (c) 2008, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.web.actions.utils;

import org.wso2.registry.*;
import org.wso2.registry.users.UserRealm;
import org.wso2.registry.exceptions.RegistryException;
import org.wso2.registry.session.UserRegistry;

import java.sql.Timestamp;
import java.util.*;

/**
 * Represents an advanced query on normal resources. Handles all details of defining, executing and
 * parameter processing of such queries.
 */
public class AdvancedResourceQuery {

    /**
     * List of parameters for the query. NOTE THAT THIS LIST SHOULD ALWAYS BE PROCESSED IN THE ORDER
     * THEY ARE LISTED BELOW.
     */
    private String resourceName;
    private String authorName;
    private String updaterName;
    private Date createdAfter;
    private Date createdBefore;
    private Date updatedAfter;
    private Date updatedBefore;
    private String commentWords;
    private String tag1;
    private String tag2;
    private String tag3;

    private String tags;
    private String queryPath;

    public Resource execute(Registry registry) throws RegistryException {

        // find if query is already defined
        // if defined, execute
        // else, create new query and execute

        String queryPath = computeQueryPath();
        if (!queryExists(queryPath)) {
            defineQuery();
        }

        List params = new ArrayList();
        if (resourceName != null && resourceName.length() > 0) {
            params.add("%/" + resourceName);
        }

        if (authorName != null && authorName.length() > 0) {
            params.add(authorName);
        }

        if (updaterName != null && updaterName.length() > 0) {
            params.add(updaterName);
        }

        if (createdAfter != null) {
            params.add(new Timestamp(createdAfter.getTime()));
        }

        if (createdBefore != null) {
            params.add(new Timestamp(createdBefore.getTime()));
        }

        if (updatedAfter != null) {
            params.add(new Timestamp(updatedAfter.getTime()));
        }

        if (updatedBefore != null) {
            params.add(new Timestamp(updatedBefore.getTime()));
        }

        if (commentWords != null && commentWords.length() > 0) {
            params.add("%" + commentWords + "%");
        }

        if (tag1 != null && tag1.length() > 0) {
            params.add(tag1);
        }

        if (tag2 != null && tag2.length() > 0) {
            params.add(tag2);
        }

        if (tag3 != null && tag3.length() > 0) {
            params.add(tag3);
        }

        Map paramMap = new HashMap();
        for (int i = 0; i < params.size(); i++) {
            Object value = params.get(i);
            paramMap.put(Integer.toString(i + 1), value);
        }

        return registry.executeQuery(queryPath, paramMap);
    }

    public String getResourceName() {
        return resourceName;
    }

    public void setResourceName(String resourceName) {
        this.resourceName = resourceName;
    }

    public String getAuthorName() {
        return authorName;
    }

    public void setAuthorName(String authorName) {
        this.authorName = authorName;
    }

    public String getUpdaterName() {
        return updaterName;
    }

    public void setUpdaterName(String updaterName) {
        this.updaterName = updaterName;
    }

    public Date getCreatedAfter() {
        return createdAfter;
    }

    public void setCreatedAfter(Date createdAfter) {
        this.createdAfter = createdAfter;
    }

    public Date getCreatedBefore() {
        return createdBefore;
    }

    public void setCreatedBefore(Date createdBefore) {
        this.createdBefore = createdBefore;
    }

    public Date getUpdatedAfter() {
        return updatedAfter;
    }

    public void setUpdatedAfter(Date updatedAfter) {
        this.updatedAfter = updatedAfter;
    }

    public Date getUpdatedBefore() {
        return updatedBefore;
    }

    public void setUpdatedBefore(Date updatedBefore) {
        this.updatedBefore = updatedBefore;
    }

    public String getTags() {
        return tags;
    }

    public void setTags(String tags) {
        this.tags = tags;

        String[] tagParts = tags.trim().split(" ");
        if (tagParts.length > 0) {
            tag1 = tagParts[0];
        }

        if (tagParts.length > 1) {
            tag2 = tagParts[1];
        }

        if (tagParts.length > 2) {
            tag3 = tagParts[2];
        }
    }

    public String getCommentWords() {
        return commentWords;
    }

    public void setCommentWords(String commentWords) {
        this.commentWords = commentWords;
    }

    private boolean queryExists(String queryPath) throws RegistryException {

        UserRegistry systemRegistry =
                (UserRegistry) System.getProperties().get(RegistryConstants.SYSTEM_REGISTRY);

        return systemRegistry.resourceExists(queryPath);
    }

    private void defineQuery() throws RegistryException {

        UserRegistry systemRegistry =
                (UserRegistry) System.getProperties().get(RegistryConstants.SYSTEM_REGISTRY);

        String queryContent = generateSQL();

        Resource q1 = systemRegistry.newResource();
        q1.setContent(queryContent);
        q1.setMediaType(RegistryConstants.SQL_QUERY_MEDIA_TYPE);
        q1.addProperty(RegistryConstants.RESULT_TYPE_PROPERTY_NAME,
                RegistryConstants.RESOURCES_RESULT_TYPE);
        String qPath = computeQueryPath();
        systemRegistry.put(qPath, q1);
    }

    private String computeQueryPath() {

        if (queryPath == null) {

            StringBuffer buf = new StringBuffer("/system/queries/advanced");

            if (resourceName != null && resourceName.length() > 0) {
                buf.append("1");
            } else {
                buf.append("0");
            }

            if (authorName != null && authorName.length() > 0) {
                buf.append("1");
            } else {
                buf.append("0");
            }

            if (updaterName != null && updaterName.length() > 0) {
                buf.append("1");
            } else {
                buf.append("0");
            }

            if (createdAfter != null) {
                buf.append("1");
            } else {
                buf.append("0");
            }

            if (createdBefore != null) {
                buf.append("1");
            } else {
                buf.append("0");
            }

            if (updatedAfter != null) {
                buf.append("1");
            } else {
                buf.append("0");
            }

            if (updatedBefore != null) {
                buf.append("1");
            } else {
                buf.append("0");
            }

            if (commentWords != null && commentWords.length() > 0) {
                buf.append("1");
            } else {
                buf.append("0");
            }

            if (tag1 != null && tag1.length() > 0) {
                buf.append("1");
            } else {
                buf.append("0");
            }

            if (tag2 != null && tag2.length() > 0) {
                buf.append("1");
            } else {
                buf.append("0");
            }

            if (tag3 != null && tag3.length() > 0) {
                buf.append("1");
            } else {
                buf.append("0");
            }

            queryPath = buf.toString();
        }

        return queryPath;
    }

    private String generateSQL() {

        boolean queryStarted = false;

        String pathQuery = "";
        if (resourceName != null && resourceName.length() != 0) {
            if (queryStarted) {
                pathQuery = " AND R.PATH LIKE ?";
            } else {
                pathQuery = " WHERE R.PATH LIKE ?";
                queryStarted = true;
            }
        }

        String authorQuery = "";
        if (authorName != null && authorName.length() != 0) {
            if (queryStarted) {
                authorQuery = " AND R.CREATOR=?";
            } else {
                authorQuery = " WHERE R.CREATOR=?";
                queryStarted = true;
            }
        }

        String updaterNameQuery = "";
        if (updaterName != null && updaterName.length() != 0) {
            if (queryStarted) {
                updaterNameQuery = " AND R.LAST_UPDATOR=?";
            } else {
                updaterNameQuery = " WHERE R.LAST_UPDATOR=?";
                queryStarted = true;
            }
        }

        String createdAfterQuery = "";
        if (createdAfter != null) {
            if (queryStarted) {
                createdAfterQuery = " AND R.CREATED_TIME>?";
            } else {
                createdAfterQuery = " WHERE R.CREATED_TIME>?";
                queryStarted = true;
            }
        }

        String createdBeforeQuery = "";
        if (createdBefore != null) {
            if (queryStarted) {
                createdBeforeQuery = " AND R.CREATED_TIME<?";
            } else {
                createdBeforeQuery = " WHERE R.CREATED_TIME<?";
                queryStarted = true;
            }
        }

        String updatedAfterQuery = "";
        if (updatedAfter != null) {
             if (queryStarted) {
                updatedAfterQuery = " AND R.LAST_UPDATED_TIME>?";
            } else {
                updatedAfterQuery = " WHERE R.LAST_UPDATED_TIME>?";
                queryStarted = true;
            }
        }

        String updatedBeforeQuery = "";
        if (updatedBefore != null) {
            if (queryStarted) {
                updatedBeforeQuery = " AND R.LAST_UPDATED_TIME<?";
            } else {
                updatedBeforeQuery = " WHERE R.LAST_UPDATED_TIME<?";
                queryStarted = true;
            }
        }

        String fromComments = "";
        String commentsQuery = "";
        if (commentWords != null && commentWords.length() != 0) {
            fromComments = ", COMMENT C";
            if (queryStarted) {
                commentsQuery = " AND C.RID=R.RID AND C.COMMENT_TEXT LIKE ?";
            } else {
                commentsQuery = " WHERE C.RID=R.RID AND C.COMMENT_TEXT LIKE ?";
                queryStarted = true;
            }
        }

        String fromTags = "";
        String tagsQuery = "";
        if (tag1 != null && tag1.length() > 0) {
            String[] tag = tags.trim().split(" ");
            if (tag.length > 0) {
                fromTags = ", TAG T";

                // todo: complete with OR tags
                if (queryStarted) {
                    tagsQuery = " AND T.RID=R.RID AND";
                } else {
                    tagsQuery = " WHERE T.RID=R.RID AND";
                    queryStarted = true;
                }

                tagsQuery = tagsQuery + " (T.TAG_NAME=?";
                for (int i = 1; i < tag.length; i++) {
                    tagsQuery = tagsQuery + " OR T.TAG_NAME=?";
                }
                tagsQuery = tagsQuery + ")";
            }
        }
       return "SELECT R.PATH FROM RESOURCE R" + fromComments +
               fromTags + pathQuery + authorQuery +
               updaterNameQuery + createdAfterQuery + createdBeforeQuery +
               updatedAfterQuery + updatedBeforeQuery + commentsQuery + tagsQuery;
    }
}
